(* July 29 last modified August 2, 2002 *)
(* Generate constraits in ALFP for RD analysis of the WHILE languge *)

(* datatype definition for the WHILE language*)
type lab = string

datatype aexpr = CstR   of real
               | Var of string           
               | Plus   of aexpr * aexpr
               | Minus  of aexpr * aexpr
               | Times  of aexpr * aexpr
               | Divide of aexpr * aexpr
          
datatype bexpr = CstB of bool
               | LES of aexpr * aexpr
               | EQ of aexpr * aexpr
               | GRT of aexpr * aexpr 
               | AND of bexpr * bexpr
               | OR  of bexpr * bexpr
               | NOT of bexpr

datatype lstmt = LSKIP  of lab 
               | LASGN  of string * aexpr * lab
               | LIF    of bexpr * lab * lstmt * lstmt 
               | LWHILE of bexpr * lab * lstmt
               | LCOMP   of lstmt * lstmt 

datatype bolck = BSKIP of lab 
               | BASGN of string * aexpr * lab
               | BBEXP of bexpr * lab  

datatype stmt =  Skip 
               | Asgn  of string * aexpr
               | If    of bexpr  * stmt * stmt 
               | While of bexpr * stmt 
               | Comp   of stmt * stmt  

exception Problem of string 


(* create a new label *) 
val blab = ref 0;

fun new() = let val _ = blab := !blab + 1
            in "L"^Int.toString(!blab)
            end

(* add lables to each of the blocks *)
fun AddLabel (Skip) = LSKIP(new())
  | AddLabel (Asgn(v, a)) = LASGN (v, a, new ())
  | AddLabel (If (b, s1, s2)) = LIF (b, new (), AddLabel (s1), AddLabel(s2))
  | AddLabel (While (b, s)) = LWHILE (b, new (), AddLabel(s))
  | AddLabel (Comp (s1, s2)) = LCOMP (AddLabel(s1), AddLabel(s2))


(* return the initial label of a statement *)
fun Init(LSKIP(l)) = l 
  | Init(LASGN(_, _, l))=l 
  | Init(LIF(_, l, _, _)) = l
  | Init(LWHILE(_, l, _)) = l
  | Init(LCOMP(s1, s2)) = Init (s1) 


(* return the set of final labels of a statement *)
fun Final(LSKIP(l)) = [l] 
  | Final(LASGN(_, _, l))=[l] 
  | Final(LIF(_, l, s1, s2)) = Final(s1) @ Final(s2) 
  | Final(LWHILE(_, l, _)) = [l]
  | Final(LCOMP(s1, s2)) = Final (s2) 


(* return the set of final labels of a statement *)
fun Blocks(LSKIP(l)) = [BSKIP(l)] 
  | Blocks(LASGN(v, a, l))=[BASGN(v,a,l)] 
  | Blocks(LIF(b, l, s1, s2)) = [BBEXP(b,l)]@Blocks(s1)@Blocks(s2) 
  | Blocks(LWHILE(b, l, s)) = [BBEXP(b,l)]@Blocks(s)
  | Blocks(LCOMP(s1, s2)) = Blocks(s1)@Blocks(s2) 


(* map statements to sets of flows *)
fun Flow (LSKIP(_)) = [] 
  | Flow(LASGN(_, _, _))=[] 
  | Flow(LIF(_, l, s1, s2)) = Flow(s1) @ Flow(s2) @ [(l, Init(s1)),(l, Init(s2))]  
  | Flow(LWHILE(_, l, s)) = Flow (s) @ [(l, Init(s))] @ (let val sr = Final(s) in map (fn x => (x, l)) sr end) 
  | Flow(LCOMP(s1, s2)) = Flow (s1) @Flow (s2) @ (let val sr = Final(s1) in map (fn x => (x, Init(s2))) sr end) 


(* return a set of paires of variables and labels of assignments that are killed by the block *)
fun FromBlosToAssLabs [] av = []
  | FromBlosToAssLabs (x::s) av = 
     case x of 
          BASGN (v, _, l) => if (v=av) then l::FromBlosToAssLabs s av else FromBlosToAssLabs s av
        | _ => FromBlosToAssLabs s av
(* s: set of blocks *)                               
fun Kill (BSKIP(_), _) = [] 
  | Kill (BBEXP(_, _), _)=[] 
  | Kill (BASGN(v, a, l), s) = [(v, "labQ" :lab)] @ (let val sr = FromBlosToAssLabs s v in map (fn x => (v, x)) sr end)
   
(* return a set of paires of variables and labels of assignments that are generated by the block *)
fun Gen (BSKIP(_)) = [] 
  | Gen (BBEXP(_, _))=[] 
  | Gen (BASGN(v, a, l)) = [(v, l)] 

(* test for a member of a set *)
fun mem a [] = false       
  | mem a (x::s) = if (a=x) then true else mem a s 

(* insert an element into a set if it is not a member of the set *)
fun insertl a s = if mem a s then s else a::s 

(* concatenate two lists with distinct elements *)
infix 6 % 
fun [] % x = x
  |  x % [] = x
  | [a] % x = insertl a x
  | x % [a] = insertl a x
  |  x % y = x @ y 

(* find free varables in an aexpr *) 
fun FvAexp (Plus(a, b)) = FvAexp a % FvAexp b
  | FvAexp (Minus(a, b)) = FvAexp a % FvAexp b
  | FvAexp (Times(a, b)) = FvAexp a % FvAexp b
  | FvAexp (Var x) = [x]
  | FvAexp _ = [] 

(* find free varables in an bexpr *) 
fun FvBexp (LES(a, b)) = FvAexp a % FvAexp b
  | FvBexp (EQ(a, b)) = FvAexp a % FvAexp b
  | FvBexp (GRT(a, b)) = FvAexp a % FvAexp b
  | FvBexp (AND(a, b)) = FvBexp a % FvBexp b
  | FvBexp (OR(a, b)) = FvBexp a % FvBexp b
  | FvBexp (NOT a) = FvBexp a 
  | FvBexp _ = [] 

(* produce a set of free variables from the set of the blocks *)
fun Fvar [] = [] 
  | Fvar (x::s) = 
         case x of (BSKIP(l)) => Fvar s
                 | (BASGN(v, a, l)) => (let val sr = FvAexp a % Fvar s in insertl v sr end)
                 | (BBEXP(b, l)) => FvBexp b % Fvar s


(* the followings generate the constraints in terms of ALFP *) 

(* add e between two elements *) 
fun between [] e = []
  | between [t] e = [t]  
  | between (x::xs) e = x::e::between xs e

(* define predicate FVAR from the set of blocks *)
fun GenFVAR p = case Fvar p of [] => raise Problem "0"
                                      |   x => String.concat (between (List.map (fn l => "FVAR("^l^")" ) x) " & " )
(* define predicate INIT  *)
fun GenINIT p = let val l = Init p in "INIT("^l^")" end

(* define predicate FLOW  *)
fun GenFLOW p = case Flow p of [] => ""
                     | [a] => ""
                     | x => String.concat (between (List.map (fn (l1, l2) => "FLOW("^l1^"," ^l2^")" ) x) " & " ) 

(* define predicate RDKILL  *)
fun GenRDKILL1 [] bs = [] 
  | GenRDKILL1 (x::xs) bs = case x of BASGN(v, a, l) => 
                                (let val s = Kill (x, bs) in map (fn (y, l1) => (l, y, l1)) s end) @ GenRDKILL1 xs bs
                             | _ => GenRDKILL1 xs bs  

fun GenRDKILL p = case GenRDKILL1 p p of [] => "" 
                | x => String.concat (between (List.map (fn (l, v, l1) => "RDKILL("^l^","^v^"," ^l1^")" ) x) " & " )

(* define predicate RDGEN  *)
fun GenRDGEN1 [] = [] 
  | GenRDGEN1 (x::xs) = case x of BASGN(v, a, l) => 
                                (let val s = Gen x in map (fn (y, l1) => (l, y, l1)) s end) @ GenRDGEN1 xs
                              | _ => GenRDGEN1 xs  

fun GenRDGEN p = case GenRDGEN1 p of [] => ""  
                   | x => String.concat (between (List.map (fn (l, v, l1) => "RDGEN("^l^","^v^"," ^l1^")" ) x) " & " )

(* generate the constraits in ALFP for a given prog *)
fun GenConst pl = (*let val pl = AddLabel prog in *)
                    let val pb = Blocks pl in 
                 GenINIT pl ^" & "^GenFVAR pb^" & "^GenFLOW pl^" & "^GenRDKILL pb^" & "^GenRDGEN pb^" & "^
                 "(A l. INIT(l) => (A x. FVAR(x) => RDIN(l, x, labQ )))"^" & "^
                 "(A l. A x. A la1. A la2. (! INIT(l) => ( RDOUT(la1, x, la2) & FLOW(la1, l) => RDIN(l, x, la2)))) & "^
                 "(A l. A x. A la1. (RDIN(l, x, la1) & (! RDKILL(l, x, la1)) | RDGEN (l, x, la1 )) => RDOUT(l, x, la1))"
                (* "(A l. A x. A la1. RDGEN(l, x, la1) => RDOUT(l, x, la1))" *) end (*end*)

(* output the costraints into a file *)
fun outputfile (data, filename) = 
    let val out = TextIO.openOut(filename) in
                  TextIO.output(out, data);
                  TextIO.closeOut(out) end;





